import scipy.optimize as opt
from multiprocessing import Pool
from sklearn.metrics import r2_score
from sklearn.linear_model import LinearRegression
import numpy as np
import uuid
import subprocess
import shlex
import os

###############################################################
#                         FUNCTIONS                           #
###############################################################


def read_file(file_name):

    open_file = open(file_name, 'r')
    data = open_file.read()
    open_file.close()
    return data


def write_file(data, file_name):

    open_file = open(file_name, 'a')
    open_file.write(data)
    open_file.close()
    return data


def extract_ref_energy(file_name):
    
    data = read_file(file_name).split('\n')
    energy = [line.split() for line in data if not (line.isspace()or line == '')]
    energy.sort(key = lambda x : x[0])
    energy = {items[0]: float(items[1]) for items in energy}
    return energy


def pars_potentials(data):

    inds_1 = [i for i, c in enumerate(data) if c == '[']
    inds_2 = [i for i, c in enumerate(data) if c == ']']
    if len(inds_1) != len(inds_2):
        raise Exception("Lost bracket in force field parameters!")
    d  = [(i_1 - (2 * i), inds_2[i] - (2 * i + 1)) for i, i_1 in enumerate(inds_1)]
    potentials = data.replace('[', '').replace(']', '')
    parameters = [float(potentials[i_1: i_2]) for i_1, i_2 in d]
    positions = [2 * i + 1 for i in range(len(d))]
    template = [potentials[0 : d[0][0]]]
    for i in range(len(d) - 1):
        template += [potentials[d[i][0]: d[i][1]]]
        template += [potentials[d[i][1]: d[i + 1][0]]]
    template += [potentials[d[-1][0]: d[-1][1]]]
    template += [potentials[d[-1][1]: len(potentials)]]
    return parameters, template, positions


def read_config():

    data = read_file("config.txt").split('\n')
    run_command = data[0].split('=')[1].strip()
    path_to_input = data[2].split('=')[1].strip()
    ref_energy = extract_ref_energy(data[4].split('=')[1].strip())
    calc_param = data[6].split('=')[1].strip()
    n_process = int(data[8].split('=')[1].strip())
    pot_param, pot_template, positions = pars_potentials('\n'.join(data[11:]))
    return run_command, path_to_input, ref_energy, calc_param, n_process, pot_param, pot_template, positions


def get_potentials(parameters, template, positions):

    potentials = template[:]
    for i, p in enumerate(parameters):
        potentials[positions[i]] = str(p)
    potentials = ''.join(potentials)
    return potentials

    
def add_calc_param(data, calc_param, functional, file_name, out_format=None):

    gin = calc_param + '\n' + data + '\n' + functional
    if out_format:
        gin += "\noutput " + out_format + ' ' + file_name
    return gin


def calc_r2(val, target):

    val = np.array(val).reshape(-1, 1)
    linmodel = LinearRegression()
    linmodel.fit(val, target)
    return r2_score(target, linmodel.predict(val))


def run_programm(gin, run_command):
    
    cwd = os.getcwd()
    temp_dir = os.path.join(cwd,  "temp/" + str(uuid.uuid4()))
    os.mkdir(temp_dir)
    os.chdir(temp_dir)
    args = shlex.split(run_command)
    p = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    result = p.communicate(gin)[0]
    os.chdir(cwd)
    os.rmdir(temp_dir)
    return result


def run_multiprocess(function, args, n_process):
    
    p = Pool(n_process)
    results = p.map(function, args)
    return results
    

def extract_gulp_energy(data):

    atom_count = None
    energy = None
    
    for line in data.split('\n'):
        if "atoms/shells" in line:
            atom_count = int(line.split('=')[1].rstrip())
        if "Total lattice energy" in line and "eV" in line:
            energy = float(line.split('=')[1].rstrip().split()[0])
    if not atom_count:
        raise Exception("The numbers of atoms is not found!")
    if not energy != None:
        raise Exception("The energy is not found!")
    return energy/atom_count


def calc_error(pot_param):

    global ref_energy
    global file_names
    global structures_data
    global iteration
    global calc_param
    global n_process
    global global_error
    global template
    global positions
    
    list_gin = []
    energy_gulp = []
    ref_energy_local = []
    log = "Iteration: " + str(iteration) + '\n'
    potentials = get_potentials(pot_param, template, positions)
    for i, data in enumerate(structures_data):
        out_path = os.path.join(os.path.join(os.getcwd(), "out"), file_names[i] + "_out")
        gin = add_calc_param(data, calc_param, potentials, out_path, out_format="cif")
        #print(gin)
        list_gin += [gin]
    gouts = run_multiprocess(run_gulp, list_gin, n_process)
    for i, gout in enumerate(gouts):
        #print(gout)
        try:
            energy_gulp += [extract_gulp_energy(gout)]
            ref_energy_local += [ref_energy[file_names[i]]]
            log += (file_names[i] + ' ' + str(extract_gulp_energy(gout)) + ' ' + str(ref_energy[file_names[i]]) + '\n')
        except:
            return 1000
    error = (1.0 - calc_r2(ref_energy_local, energy_gulp)) * 100
    log += "Error: " + str(error) + " Potential paramiters: " + ''.join([str(p) + ' ' for p in pot_param]) + '\n\n'
    if global_error > error:
        global_error = error
        write_file(log, 'log.txt')
        print("Error: " + str(error) + " Potential paramiters: " + ''.join([str(p) + ' ' for p in pot_param]))
        iteration += 1
    return error

###############################################################
#                    START OF THE SCRIPT                      #
###############################################################


if __name__ == "__main__":
    iteration = 1
    global_error = float('inf')
    methods = ['BFGS','Newton-CG','Anneal','L-BFGS-B','TNC','COBYLA','SLSQP','dogleg','trust-ncg']
    run_command, path_to_input, ref_energy, calc_param, n_process, pot_param, template, positions = read_config()
    file_names = [f[: -4] for f in os.listdir(path_to_input) if f.endswith('.gin') and ref_energy.get(f[: -4]) is not None]
    structures_data = [read_file(os.path.join(path_to_input, f + '.gin')) for f in file_names]
    if "temp" not in os.listdir('.'):
        os.mkdir("temp")
    if "out" not in os.listdir('.'):
        os.mkdir("out")
            
    def run_gulp(gin):

        global run_command
        return run_programm(gin, run_command)
    
    
    param = opt.minimize(calc_error, pot_param, method=methods[5]).x
    print("Final potential paramiters: " + ''.join([str(p) + ' ' for p in param]))

    
